/*  ----------------------------------------------------------------------

    Copyright (C) 2000  Gabriel Gomiz  (gomita@bblanca.com.ar)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    ---------------------------------------------------------------------- */

    /* $Id$ */

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include "nomadii.h"

int     from_a_tty;
int	memory_bank;
char	cmdline[200];
int	margc;
char	*margv[200];
char	*prog_name;

nomadII_device nomadII;

static void cmdscanner();
static void makeargv();

static struct option const long_options[] =
{
  {"version",     no_argument,       0,    'V' },
  {"help",        no_argument,       0,    'h' },
  {"autorename",  no_argument,       0,    'a' },
  {"verbose",     no_argument,       0,    'v' },
  {"bell",        no_argument,       0,    'b' },
  {"hash",        no_argument,       0,    'H' },
  {"prompt",      no_argument,       0,    't' },
  {"force",       no_argument,       0,    'F' },
  {"probe",       no_argument,       0,    'P' },
  {"model",       no_argument,       0,    'M' },
  {"memory",      no_argument,       0,    'm' },
  {"count",       no_argument,       0,    'c' },
  {"list",        no_argument,       0,    'l' },
  {"format",      no_argument,       0,    'f' },
  {"internal",    no_argument,       0,    'i' },
  {"external",    no_argument,       0,    'x' },
  {"replace",     no_argument,       0,    'r' },
  {"delete",      required_argument, 0,    'D' },
  {"send",	  required_argument, 0,    's' },
  {"put",	  required_argument, 0,    's' },
  {"download",    required_argument, 0,    'd' },
  {"playlist",    required_argument, 0,    'p' },
  {"get",         required_argument, 0,    'g' },
  {"upload",      required_argument, 0,    'u' },
  {"offset",      required_argument, 0,    'o' },  
  {"number",	  required_argument, 0,    'n'  },
  {"freq",        required_argument, 0,    OPTION_FREQ },
  {"clock-set",   no_argument,       0,    OPTION_CLOCK_SET  },
  {"clock",       no_argument,       0,    OPTION_CLOCK  },
  {"owner-set",   required_argument, 0,    OPTION_OWNER_SET  },
  {"owner",       no_argument,       0,    OPTION_OWNER  },
  {"presets-set", required_argument, 0,    OPTION_PRESETS_SET  },
  {"presets",     no_argument,       0,    OPTION_PRESETS  },

#ifdef PACKET_DUMP
  {"debug",       no_argument,       0,    OPTION_DEBUG  },
#endif

  {NULL,          0,                 NULL, 0   }
};

int autorename;
int interactive;
int verbose;
int bell;
int hash;
int prompt;
int force;

#ifdef PACKET_DUMP
int debug;
#endif

/* 
 * Function prototypes 
 */
struct	cmd *getcmd();
void intr();

int
main (int argc, char *argv[])
{
	int c, i, number;
	unsigned int offset,freq,subfreq,replace; 
	
	/* done as early as possible to shed setuid */
	dev_probe();

	nomadII_set_progress_handler(progress_handler, NULL, 48);

#ifdef USE_READLINE
	from_a_tty = isatty(fileno(stdin));
        rl_terminal_name = (char *) getenv("TERM");
#endif

	/* defaults */
	verbose = 0;
	autorename = 1;
	interactive = 0;
	bell = 0;
	hash = 1;
	prompt = 0;
	force = 0;
	offset = 0;
	freq = 0;
	subfreq = 0;
	replace = 0;
	memory_bank = BOTH;
	prog_name   = strrchr(argv[0], '/');
	number = -1;

#ifdef PACKET_DUMP
	debug = 0;
#endif

	if (prog_name == NULL)
		prog_name = argv[0];
	else
		prog_name++;

	/* Command line options */

	while ((c = getopt_long (argc, argv,
				 "abcs:d:fhitxlmn:o:p:rg:u:vD:FHPMV", long_options, NULL)) != -1)
	{
		switch (c)
		{
		case 0:
			break;
			
		case 'v':
			verbose = 1;
			fprintf(stderr, "verbose\n");
			break;

		case 'a':
			autorename = 0;
			break;

		case 'b':
			bell = 1;
			break;

		case 'H':
			hash = 0;
			break;

		case 't':
			prompt = 1;
			break;

		case 'F':
			force = 1;
			break;
		case 'i': 
			/* this will not cope with multiple i/x's */
			if (memory_bank == EXTERNAL)
				memory_bank = BOTH;
			else
				memory_bank = INTERNAL;
			break;
		case 'x':
			/* this will not cope with multiple i/x's */
			if (memory_bank == INTERNAL)
				memory_bank = BOTH;
			else
				memory_bank = EXTERNAL;
			break;
		case 'o':
			offset = atoi(optarg);
			break;
		case 'n':
			number = atoi(optarg);
			break;
		case 'r':
			replace = 1;
			break;
		case OPTION_FREQ:
			freq = atoi(optarg);
			subfreq = freq % 100;
			freq /= 100;
			break;

#ifdef PACKET_DUMP
		case OPTION_DEBUG:
			nomadIIdebug++;
			debug = 1;
			break;
#endif

		case 'h':
			usage();
			exit(0);

		case 'V':
			version();
			exit(0);

		case 'l':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (list(0, NULL) != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case 'c':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (count() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case 'm':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (memory() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case 'M':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (model() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case 'f':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (format() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case 'P':
			exit(dev_probe_status(FALSE));

		case OPTION_CLOCK:
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (get_clock() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case OPTION_CLOCK_SET:
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (set_clock() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case OPTION_OWNER:
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (get_owner() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case OPTION_OWNER_SET:                
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (set_owner(optarg) != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case OPTION_PRESETS:
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (presets(0, NULL) != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);		

		case OPTION_PRESETS_SET:
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (set_preset(atoi(optarg),freq,subfreq) != 0) return -1;
			if (set_presets() != 0) return -1;
			if (dev_close() != 0) return -1;
			exit(0);

		case 's':
		case 'd':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (replace)
				if (do_delete_all() != 0) return -1;
			progress_handler(NOMADII_PROGRESS_RESET, argc - optind + 1, NULL);
			if (do_download(optarg) != 0) return -1;
			for (i = optind; i < argc; i++)
				if (*argv[i] == '-')
					break;
				else if (do_download(argv[i]) != 0) 
					return -1;
			if (nomadII_sync(&nomadII) != 0)
				fprintf(stderr,"Warning: Unable to get in sync with device: %s\n", nomadII_get_error());
			if (bell) putchar('\007');
			if (dev_close() != 0) return -1;
			exit(0);	

		case 'D':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (strcmp(optarg, "all") == 0) {
				if (do_delete_all() != 0) return -1;
			} else {
				if (do_delete(atoi(optarg))) return -1;
				for (i = optind; i < argc; i++)
					if (*argv[i] == '-')
						break;
					else if (do_delete(atoi(argv[i])) != 0) 
						return -1;
			}
			if (nomadII_sync(&nomadII) != 0)
				fprintf(stderr,"Warning: Unable to get in sync with device: %s\n", nomadII_get_error());
			if (bell) putchar('\007');
			if (dev_close() != 0) return -1;
			exit(0);

		case 'g':
		case 'u':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (strcmp(optarg, "all") == 0) {
				if (do_upload_all() != 0) return -1;
			} else {
				progress_handler(NOMADII_PROGRESS_RESET, argc - optind + 1, NULL);
				if (do_upload(atoi(optarg)) != 0) return -1;
				for (i = optind; i < argc; i++)
					if (*argv[i] == '-')
						break;
					else if (do_upload(atoi(argv[i])) != 0) 
						return -1;
			}
			if (nomadII_sync(&nomadII) != 0)
				fprintf(stderr,"Warning: Unable to get in sync with device: %s\n", nomadII_get_error());
			if (bell) putchar('\007');
			if (dev_close() != 0) return -1;
			exit(0);

		case 'p':
			if (dev_probe_status(TRUE) != 0) return -1;
			if (dev_open() != 0) return -1;
			if (replace)
				if (do_delete_all() != 0) return -1;
			if (do_playlist(optarg,offset,number) != 0) return -1;
			if (nomadII_sync(&nomadII) != 0)
				fprintf(stderr,"Warning: Unable to get in sync with device: %s\n", nomadII_get_error());
			if (bell) putchar('\007');
			if (dev_close() != 0) return -1;
			exit(0);

		default:
			usage();
			exit(-1);
		}
	}

	signal(SIGINT, intr);

	if (optind != argc) {
		fprintf(stderr, "all commands must use - or -- form\n");	
		usage();
		exit(-1);
	}

	interactive = 1;

	for (;;) 
		cmdscanner();
}

/*
 * Signal Handler - SIGINT
 */
void intr()
{
	exit(0);
}


/* this does an integer log function without -lm needed */
static int numeric_digits(int number) {
	int count = 0;

	do  {
		number /= 10;
		count++;
	} while (number > 0);
	return count;
}

void progress_handler(int message, int value, void* data)
{
	static int progress_total = 0;
	static int progress_count = 0;
	static int lastProgress   = 0;
	static int dupCRCFound    = 0;

	int i;
	
	switch (message) {
		case NOMADII_PROGRESS_NEW_FILE:
			lastProgress = 0;
			dupCRCFound  = 0;
			progress_count++;
			return;
		case NOMADII_PROGRESS_RESET:
			progress_count = 0;
			progress_total = value;
			return;
		case NOMADII_PROGRESS_DUPCRC:
			if (dupCRCFound >= 0)
				dupCRCFound++;
			else /* negative to mark a * has been printed for it */
				dupCRCFound = -dupCRCFound + 1;
			if (verbose) {
				fprintf(stdout, "<%d>", value);
			}
			return;
	}

	/* else NOMADII_PROGRESS_UPDATE */

	if (!hash) {
		if (value == 48 && dupCRCFound != 0)
			fprintf(stdout, "warning: %d duplicate block%s found and corrected\n", 
			    dupCRCFound, (dupCRCFound == 1 ? "" : "s"));
		return;
	}

	for (i=lastProgress; i <= value; i++) {
		if (i == 0)
			fprintf(stdout,"%*d/%d: [0%%", 
			    numeric_digits(progress_total), 
			    progress_count, progress_total);
		else if (i == 12)
			printf("25%%");
		else if (i == 24)
			printf("50%%");
		else if (i == 36)
			printf("75%%");
		if (i == 48) {
			printf("100%%]\n");
			if (dupCRCFound != 0)
				fprintf(stdout, "warning: %d duplicate block%s found and corrected\n", 
				    -dupCRCFound, (dupCRCFound == -1 ?  "" : "s"));
		} else {
			if (dupCRCFound > 0) {
				dupCRCFound *= -1;
				putchar('*');
			} else {
				putchar('.');
			}
		}
	}
	lastProgress = value + 1;
	fflush(stdout);
}

static char *get_input_line(char *buf, int buflen)
{
	static char *cmdprompt = NULL;
	
	if (cmdprompt == NULL) {
		cmdprompt = (char*)malloc(strlen(prog_name) + 3);
		strcpy(cmdprompt, prog_name);
		strcat(cmdprompt, "> ");
	}

#ifdef USE_READLINE
        if (from_a_tty && !rl_inhibit_completion) {
                char *lineread = readline(cmdprompt);
                if (!lineread) return NULL;
                strncpy(buf, lineread, buflen);
                buf[buflen-1] = 0;
                if (lineread[0]) add_history(lineread);
                free(lineread);
                return buf;
        }
#endif
        if (from_a_tty) {
                fprintf(stdout,cmdprompt);
                fflush(stdout);
        }
        return fgets(buf, buflen, stdin);
}

/*
 * Command parser.
 */
static void
cmdscanner()
{
	register struct cmd *c;
	register int l;

	if (verbose)
		version();

	if (dev_probe_status(FALSE) != 0)
		exit(-1);

	if (dev_open() != 0)
		exit(-1);

#ifdef USE_READLINE
	rl_filename_quote_characters = " ";
#endif

	for (;;) {
		if (!get_input_line(cmdline, sizeof(cmdline))) exit(-1);

		l = strlen(cmdline);
		if (l == 0)
			continue;
		if (cmdline[--l] == '\n') {
			if (l == 0)
				continue;
			cmdline[l] = '\0';
		} 
		else if (l == sizeof(cmdline) - 2) {
			fprintf(stderr,"sorry, input line too long\n");
			while ((l = getchar()) != '\n' && l != EOF);
			continue;
		}

		makeargv();
		c = getcmd(margv[0]);
		if (c == (struct cmd *)-1) {
			fprintf(stderr,"?Ambiguous command\n");
			continue;
		}
		if (c == 0) {
			fprintf(stderr,"?Invalid command\n");
			continue;
		}
		(*c->c_handler)(margc, margv);
	}
}


/* 
 * Command Table 
 */
extern struct cmd cmdtab[];

struct cmd *getcmd(register char *name)
{
	register char *p, *q;
	register struct cmd *c, *found;
	register int nmatches, longest;

	longest = 0;
	nmatches = 0;
	found = 0;
	for (c = cmdtab; (p = c->c_name) != NULL; c++) {
		for (q = name; *q == *p++; q++)
			if (*q == 0)		/* exact match? */
				return(c);
		if (!*q) {			/* the name was a prefix */
			if (q - name > longest) {
				longest = q - name;
				nmatches = 1;
				found = c;
			} else if (q - name == longest)
				nmatches++;
		}
	}
	if (nmatches > 1)
		return((struct cmd *)-1);
	return(found);
}

/*
 * Slice a string up into argc/argv.
 */
void
makeargv()
{
	register char *cp;
	register char **argp = margv;
	int composite;

	margc = 0;
	for (cp = cmdline; *cp;) {
		if (*cp == '"') {
			composite = 1;
			cp++;
		}
		else {
			while (isspace(*cp))
				cp++;
			composite = 0;
		}
		if (*cp == '\0')
			break;
		*argp++ = cp;
		margc += 1;
		if (composite) {
			while (*cp != '\0' && *cp != '"')
				cp++;
		}
		else {
			while (*cp != '\0' && !isspace(*cp))
				cp++;
		}
		if (*cp == '\0')
			break;
		*cp++ = '\0';
	}
	*argp++ = 0;
}
